"""
Local animations

/524 load_local_anim
 assuming you loaded something.anim
/524 start_local_anim something
/524 stop_local_anim something
/524 save_local_anim something

If you want to trigger the animation from an object to simulate llStartAnimation():
llOwnerSay("@start_local_anim:something=force");

Also includes a concept of "anim manglers" similar to the "mesh manglers" of the
local mesh addon. This is useful if you want to test making procedural changes
to animations before uploading them. The manglers will be applied to any uploaded
animations as well.

May also be useful if you need to make ad-hoc changes to a bunch of animations on
bulk upload, like changing priority or removing a joint.
"""

import asyncio
import logging
import pathlib
from abc import abstractmethod
from typing import *

from hippolyzer.lib.base import serialization as se
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.helpers import get_mtime
from hippolyzer.lib.base.llanim import Animation
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.message.msgtypes import PacketFlags
from hippolyzer.lib.proxy import addon_ctx
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.addon_utils import BaseAddon, SessionProperty, GlobalProperty, show_message
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.http_asset_repo import HTTPAssetRepo
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager


class LocalAnimAddon(BaseAddon):
    # name -> path, only for anims actually from files
    local_anim_paths: Dict[str, str] = SessionProperty(dict)
    # name -> anim bytes
    local_anim_bytes: Dict[str, bytes] = SessionProperty(dict)
    # name -> mtime or None. Only for anims from files.
    local_anim_mtimes: Dict[str, Optional[float]] = SessionProperty(dict)
    # name -> current asset ID (changes each play)
    local_anim_playing_ids: Dict[str, UUID] = SessionProperty(dict)
    anim_manglers: List[Callable[[Animation], Animation]] = GlobalProperty(list)

    def handle_init(self, session_manager: SessionManager):
        self.remangle_local_anims(session_manager)

    def handle_session_init(self, session: Session):
        # Reload anims and reload any manglers if we have any
        self._schedule_task(self._try_reload_anims(session))

    @handle_command()
    async def load_local_anim(self, _session: Session, _region: ProxiedRegion):
        """Load a local animation file into the list of local anims"""
        filename = await AddonManager.UI.open_file(filter_str="SL Anim (*.anim)")
        if filename:
            p = pathlib.Path(filename)
            self.local_anim_paths[p.stem] = filename

    @handle_command(anim_name=str)
    async def start_local_anim(self, session: Session, region: ProxiedRegion, anim_name):
        """
        Start a named local animation

        Assuming you loaded an animation named something.anim:
            start_local_anim something
        """
        self.apply_local_anim_from_file(session, region, anim_name)

    @handle_command(anim_name=str)
    async def stop_local_anim(self, session: Session, region: ProxiedRegion, anim_name):
        """Stop a named local animation"""
        self.apply_local_anim(session, region, anim_name, new_data=None)

    @handle_command(anim_name=str)
    async def save_local_anim(self, _session: Session, _region: ProxiedRegion, anim_name: str):
        """Save a named local anim to disk"""
        anim_bytes = self.local_anim_bytes.get(anim_name)
        if not anim_bytes:
            return
        filename = await AddonManager.UI.save_file(filter_str="SL Anim (*.anim)", default_suffix="anim")
        if not filename:
            return
        with open(filename, "wb") as f:
            f.write(anim_bytes)

    async def _try_reload_anims(self, session: Session):
        while True:
            region = session.main_region
            if not region:
                await asyncio.sleep(1.0)
                continue

            # Loop over local anims we loaded
            for anim_name, anim_id in self.local_anim_paths.items():
                anim_id = self.local_anim_playing_ids.get(anim_name)
                if not anim_id:
                    continue
                # is playing right now, check if there's a newer version
                try:
                    self.apply_local_anim_from_file(session, region, anim_name, only_if_changed=True)
                except Exception:
                    logging.exception("Exploded while replaying animation")
            await asyncio.sleep(1.0)

    def handle_rlv_command(self, session: Session, region: ProxiedRegion, source: UUID,
                           behaviour: str, options: List[str], param: str):
        # We only handle commands
        if param != "force":
            return

        if behaviour == "stop_local_anim":
            self.apply_local_anim(session, region, options[0], new_data=None)
            return True
        elif behaviour == "start_local_anim":
            self.apply_local_anim_from_file(session, region, options[0])
            return True

    @classmethod
    def apply_local_anim(cls, session: Session, region: ProxiedRegion,
                         anim_name: str, new_data: Optional[bytes] = None):
        asset_repo: HTTPAssetRepo = session.session_manager.asset_repo
        next_id: Optional[UUID] = None
        new_msg = Message(
            "AgentAnimation",
            Block(
                "AgentData",
                AgentID=session.agent_id,
                SessionID=session.id,
            ),
            flags=PacketFlags.RELIABLE,
        )

        # Stop any old version of the anim that might be playing first
        cur_id = cls.local_anim_playing_ids.get(anim_name)
        if cur_id:
            new_msg.add_block(Block(
                "AnimationList",
                AnimID=cur_id,
                StartAnim=False,
            ))

        if new_data is not None:
            # Create a temp asset ID for the new version and send out a start request
            next_id = asset_repo.create_asset(new_data, one_shot=True)
            new_msg.add_block(Block(
                "AnimationList",
                AnimID=next_id,
                StartAnim=True,
            ))
            cls.local_anim_playing_ids[anim_name] = next_id
            cls.local_anim_bytes[anim_name] = new_data
        else:
            # No data means just stop the anim
            cls.local_anim_playing_ids.pop(anim_name, None)
            cls.local_anim_bytes.pop(anim_name, None)

        region.circuit.send(new_msg)
        print(f"Changing {anim_name} to {next_id}")

    @classmethod
    def apply_local_anim_from_file(cls, session: Session, region: ProxiedRegion,
                                   anim_name: str, only_if_changed=False):
        anim_path = cls.local_anim_paths.get(anim_name)
        anim_data = None
        if anim_path:
            old_mtime = cls.local_anim_mtimes.get(anim_name)
            mtime = get_mtime(anim_path)
            if only_if_changed and old_mtime == mtime:
                return

            # file might not even exist anymore if mtime is `None`,
            # anim will automatically stop if that happens.
            if mtime:
                if only_if_changed:
                    print(f"Re-applying {anim_name}")
                else:
                    print(f"Playing {anim_name}")

                with open(anim_path, "rb") as f:
                    anim_data = f.read()
                anim_data = cls._mangle_anim(anim_data)
            cls.local_anim_mtimes[anim_name] = mtime
        else:
            print(f"Unknown anim {anim_name!r}")
        cls.apply_local_anim(session, region, anim_name, new_data=anim_data)

    @classmethod
    def _mangle_anim(cls, anim_data: bytes) -> bytes:
        if not cls.anim_manglers:
            return anim_data
        reader = se.BufferReader("<", anim_data)
        spec = se.Dataclass(Animation)
        anim = reader.read(spec)
        for mangler in cls.anim_manglers:
            anim = mangler(anim)
        writer = se.BufferWriter("<")
        writer.write(spec, anim)
        return writer.copy_buffer()

    @classmethod
    def remangle_local_anims(cls, session_manager: SessionManager):
        # Anim manglers are global, so we need to re-mangle anims for all sessions
        for session in session_manager.sessions:
            # Push the context of this session onto the stack so we can access
            # session-scoped properties
            with addon_ctx.push(new_session=session, new_region=session.main_region):
                cls.local_anim_mtimes.clear()

    def handle_http_request(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        if flow.name == "NewFileAgentInventoryUploader":
            # Don't bother looking at this if we have no manglers
            if not self.anim_manglers:
                return
            # This is kind of a crappy match but these magic bytes shouldn't match anything that SL
            # allows as an upload type but animations.
            if not flow.request.content or not flow.request.content.startswith(b"\x01\x00\x00\x00"):
                return

            # Replace the uploaded anim with the mangled version
            flow.request.content = self._mangle_anim(flow.request.content)
            show_message("Mangled upload request")


class BaseAnimManglerAddon(BaseAddon):
    """Base class for addons that mangle uploaded or file-based local animations"""
    ANIM_MANGLERS: List[Callable[[Animation], Animation]]

    def handle_init(self, session_manager: SessionManager):
        # Add our manglers into the list
        LocalAnimAddon.anim_manglers.extend(self.ANIM_MANGLERS)
        LocalAnimAddon.remangle_local_anims(session_manager)

    def handle_unload(self, session_manager: SessionManager):
        # Clean up our manglers before we go away
        mangler_list = LocalAnimAddon.anim_manglers
        for mangler in self.ANIM_MANGLERS:
            if mangler in mangler_list:
                mangler_list.remove(mangler)
        LocalAnimAddon.remangle_local_anims(session_manager)


class BaseAnimHelperAddon(BaseAddon):
    """
    Base class for local creation of procedural animations

    Animation generated by build_anim() gets applied to all active sessions
    """
    ANIM_NAME: str

    def handle_session_init(self, session: Session):
        self._reapply_anim(session, session.main_region)

    def handle_session_closed(self, session: Session):
        LocalAnimAddon.apply_local_anim(session, session.main_region, self.ANIM_NAME, None)

    def handle_unload(self, session_manager: SessionManager):
        for session in session_manager.sessions:
            # TODO: Nasty. Since we need to access session-local attrs we need to set the
            #  context even though we also explicitly pass session and region.
            #  Need to rethink the LocalAnimAddon API.
            with addon_ctx.push(session, session.main_region):
                LocalAnimAddon.apply_local_anim(session, session.main_region, self.ANIM_NAME, None)

    @abstractmethod
    def build_anim(self) -> Animation:
        pass

    def _reapply_anim(self, session: Session, region: ProxiedRegion):
        LocalAnimAddon.apply_local_anim(session, region, self.ANIM_NAME, self.build_anim().to_bytes())


addons = [LocalAnimAddon()]



import unittest
from typing import Any

import aioresponses

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base import llsd
from hippolyzer.lib.client.object_manager import ClientObjectManager

from . import MockClientRegion


class MaterialManagerTest(unittest.IsolatedAsyncioTestCase):
    FAKE_CAPS = {
        "RenderMaterials": "http://127.0.0.1:8023"
    }

    GET_RENDERMATERIALS_BODY = [
        {'ID': UUID(int=1).bytes,
         'Material': {'AlphaMaskCutoff': 0, 'DiffuseAlphaMode': 1, 'EnvIntensity': 0,
                      'NormMap': UUID(int=4), 'NormOffsetX': 0, 'NormOffsetY': 0,
                      'NormRepeatX': 10000, 'NormRepeatY': 10000, 'NormRotation': 0, 'SpecColor': [255, 255, 255, 255],
                      'SpecExp': 51, 'SpecMap': UUID(int=5), 'SpecOffsetX': 0,
                      'SpecOffsetY': 0, 'SpecRepeatX': 10000, 'SpecRepeatY': 10000, 'SpecRotation': 0}},
        {'ID': UUID(int=2).bytes,
         'Material': {'AlphaMaskCutoff': 0, 'DiffuseAlphaMode': 0, 'EnvIntensity': 0,
                      'NormMap': UUID(int=6), 'NormOffsetX': 0, 'NormOffsetY': 0,
                      'NormRepeatX': 10000, 'NormRepeatY': -10000, 'NormRotation': 0,
                      'SpecColor': [255, 255, 255, 255], 'SpecExp': 51,
                      'SpecMap': UUID(int=7), 'SpecOffsetX': 0, 'SpecOffsetY': 0,
                      'SpecRepeatX': 10000, 'SpecRepeatY': -10000, 'SpecRotation': 0}},
        {'ID': UUID(int=3).bytes,
         'Material': {'AlphaMaskCutoff': 0, 'DiffuseAlphaMode': 1, 'EnvIntensity': 50,
                      'NormMap': UUID.ZERO, 'NormOffsetX': 0, 'NormOffsetY': 0,
                      'NormRepeatX': 10000, 'NormRepeatY': 10000, 'NormRotation': 0, 'SpecColor': [255, 255, 255, 255],
                      'SpecExp': 200, 'SpecMap': UUID(int=8), 'SpecOffsetX': 0,
                      'SpecOffsetY': 0, 'SpecRepeatX': 10000, 'SpecRepeatY': 10000, 'SpecRotation': 0}},
    ]

    def _make_rendermaterials_resp(self, resp: Any) -> bytes:
        return llsd.format_xml({"Zipped": llsd.zip_llsd(resp)})

    async def asyncSetUp(self):
        self.aio_mock = aioresponses.aioresponses()
        self.aio_mock.start()
        # Requesting all materials
        self.aio_mock.get(
            self.FAKE_CAPS['RenderMaterials'],
            body=self._make_rendermaterials_resp(self.GET_RENDERMATERIALS_BODY)
        )
        # Specific material request
        self.aio_mock.post(
            self.FAKE_CAPS['RenderMaterials'],
            body=self._make_rendermaterials_resp([self.GET_RENDERMATERIALS_BODY[0]])
        )
        self.region = MockClientRegion(self.FAKE_CAPS)
        self.manager = ClientObjectManager(self.region)

    async def asyncTearDown(self):
        self.aio_mock.stop()

    async def test_fetch_all_materials(self):
        await self.manager.request_all_materials()
        self.assertListEqual([UUID(int=1), UUID(int=2), UUID(int=3)], list(self.manager.state.materials.keys()))

    async def test_fetch_some_materials(self):
        mats = await self.manager.request_materials((UUID(int=1),))
        self.assertListEqual([UUID(int=1)], list(mats.keys()))
        self.assertListEqual([UUID(int=1)], list(self.manager.state.materials.keys()))



"""
Body parts and linden clothing layers
"""

from __future__ import annotations

import dataclasses
import enum
import logging
from io import StringIO
from typing import *

from xml.etree.ElementTree import parse as parse_etree

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.helpers import get_resource_filename
from hippolyzer.lib.base.inventory import InventorySaleInfo, InventoryPermissions
from hippolyzer.lib.base.legacy_schema import SchemaBase, parse_schema_line, SchemaParsingError
import hippolyzer.lib.base.serialization as se
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.templates import WearableType

LOG = logging.getLogger(__name__)
_T = TypeVar("_T")

WEARABLE_VERSION = "LLWearable version 22"
DEFAULT_WEARABLE_TEX = UUID("c228d1cf-4b5d-4ba8-84f4-899a0796aa97")


class AvatarTEIndex(enum.IntEnum):
    """From llavatarappearancedefines.h"""
    HEAD_BODYPAINT = 0
    UPPER_SHIRT = enum.auto()
    LOWER_PANTS = enum.auto()
    EYES_IRIS = enum.auto()
    HAIR = enum.auto()
    UPPER_BODYPAINT = enum.auto()
    LOWER_BODYPAINT = enum.auto()
    LOWER_SHOES = enum.auto()
    HEAD_BAKED = enum.auto()
    UPPER_BAKED = enum.auto()
    LOWER_BAKED = enum.auto()
    EYES_BAKED = enum.auto()
    LOWER_SOCKS = enum.auto()
    UPPER_JACKET = enum.auto()
    LOWER_JACKET = enum.auto()
    UPPER_GLOVES = enum.auto()
    UPPER_UNDERSHIRT = enum.auto()
    LOWER_UNDERPANTS = enum.auto()
    SKIRT = enum.auto()
    SKIRT_BAKED = enum.auto()
    HAIR_BAKED = enum.auto()
    LOWER_ALPHA = enum.auto()
    UPPER_ALPHA = enum.auto()
    HEAD_ALPHA = enum.auto()
    EYES_ALPHA = enum.auto()
    HAIR_ALPHA = enum.auto()
    HEAD_TATTOO = enum.auto()
    UPPER_TATTOO = enum.auto()
    LOWER_TATTOO = enum.auto()
    HEAD_UNIVERSAL_TATTOO = enum.auto()
    UPPER_UNIVERSAL_TATTOO = enum.auto()
    LOWER_UNIVERSAL_TATTOO = enum.auto()
    SKIRT_TATTOO = enum.auto()
    HAIR_TATTOO = enum.auto()
    EYES_TATTOO = enum.auto()
    LEFT_ARM_TATTOO = enum.auto()
    LEFT_LEG_TATTOO = enum.auto()
    AUX1_TATTOO = enum.auto()
    AUX2_TATTOO = enum.auto()
    AUX3_TATTOO = enum.auto()
    LEFTARM_BAKED = enum.auto()
    LEFTLEG_BAKED = enum.auto()
    AUX1_BAKED = enum.auto()
    AUX2_BAKED = enum.auto()
    AUX3_BAKED = enum.auto()

    @property
    def is_baked(self) -> bool:
        return self.name.endswith("_BAKED")


class VisualParamGroup(enum.IntEnum):
    TWEAKABLE = 0
    ANIMATABLE = 1
    TWEAKABLE_NO_TRANSMIT = 2
    TRANSMIT_NOT_TWEAKABLE = 3


@dataclasses.dataclass
class VisualParam:
    id: int
    name: str
    value_min: float
    value_max: float
    value_default: float
    group: VisualParamGroup
    # These might be `None` if the param isn't meant to be directly edited
    edit_group: Optional[str]
    wearable: Optional[str]

    def dequantize_val(self, val: int) -> float:
        """Dequantize U8 values from AvatarAppearance messages"""
        spec = se.QuantizedFloat(se.U8, self.value_min, self.value_max, False)
        return spec.decode(val, None)


class VisualParams(List[VisualParam]):
    def __init__(self, lad_path):
        super().__init__()
        with open(lad_path, "rb") as f:
            doc = parse_etree(f)

        temp_params = []
        for param in doc.findall(".//param"):
            temp_params.append(VisualParam(
                id=int(param.attrib["id"]),
                name=param.attrib["name"],
                group=VisualParamGroup(int(param.get("group", "0"))),
                edit_group=param.get("edit_group"),
                wearable=param.get("wearable"),
                value_min=float(param.attrib["value_min"]),
                value_max=float(param.attrib["value_max"]),
                value_default=float(param.attrib.get("value_default", 0.0))
            ))
        # Some functionality relies on the list being sorted by ID, though there may be holes.
        temp_params.sort(key=lambda x: x.id)
        # Remove dupes, only using the last value present (matching indra behavior)
        # This is necessary to remove the duplicate eye pop entry...
        self.extend({x.id: x for x in temp_params}.values())

    @property
    def appearance_params(self) -> Iterator[VisualParam]:
        for param in self:
            if param.group not in (VisualParamGroup.TWEAKABLE, VisualParamGroup.TRANSMIT_NOT_TWEAKABLE):
                continue
            yield param

    def by_name(self, name: str) -> VisualParam:
        return [x for x in self if x.name == name][0]

    def by_edit_group(self, edit_group: str) -> List[VisualParam]:
        return [x for x in self if x.edit_group == edit_group]

    def by_wearable(self, wearable: str) -> List[VisualParam]:
        return [x for x in self if x.wearable == wearable]

    def by_id(self, vparam_id: int) -> VisualParam:
        return [x for x in self if x.id == vparam_id][0]

    def parse_appearance_message(self, message: Message) -> Dict[int, float]:
        params = {}
        for param, value_block in zip(self.appearance_params, message["VisualParam"]):
            params[param.id] = param.dequantize_val(value_block["ParamValue"])
        return params


VISUAL_PARAMS = VisualParams(get_resource_filename("lib/base/data/avatar_lad.xml"))


# See `llpaneleditwearable.cpp`, which TE slots should be set for each wearable type is hardcoded
# in the viewer.
WEARABLE_TEXTURE_SLOTS: Dict[WearableType, Sequence[AvatarTEIndex]] = {
    WearableType.SHAPE: (),
    WearableType.SKIN: (AvatarTEIndex.HEAD_BODYPAINT, AvatarTEIndex.UPPER_BODYPAINT, AvatarTEIndex.LOWER_BODYPAINT),
    WearableType.HAIR: (AvatarTEIndex.HAIR,),
    WearableType.EYES: (AvatarTEIndex.EYES_IRIS,),
    WearableType.SHIRT: (AvatarTEIndex.UPPER_SHIRT,),
    WearableType.PANTS: (AvatarTEIndex.LOWER_PANTS,),
    WearableType.SHOES: (AvatarTEIndex.LOWER_SHOES,),
    WearableType.SOCKS: (AvatarTEIndex.LOWER_SOCKS,),
    WearableType.JACKET: (AvatarTEIndex.UPPER_JACKET, AvatarTEIndex.LOWER_JACKET),
    WearableType.GLOVES: (AvatarTEIndex.UPPER_GLOVES,),
    WearableType.UNDERSHIRT: (AvatarTEIndex.UPPER_UNDERSHIRT,),
    WearableType.UNDERPANTS: (AvatarTEIndex.LOWER_UNDERPANTS,),
    WearableType.SKIRT: (AvatarTEIndex.SKIRT,),
    WearableType.ALPHA: (AvatarTEIndex.LOWER_ALPHA, AvatarTEIndex.UPPER_ALPHA,
                         AvatarTEIndex.HEAD_ALPHA, AvatarTEIndex.EYES_ALPHA, AvatarTEIndex.HAIR_ALPHA),
    WearableType.TATTOO: (AvatarTEIndex.LOWER_TATTOO, AvatarTEIndex.UPPER_TATTOO, AvatarTEIndex.HEAD_TATTOO),
    WearableType.UNIVERSAL: (AvatarTEIndex.HEAD_UNIVERSAL_TATTOO, AvatarTEIndex.UPPER_UNIVERSAL_TATTOO,
                             AvatarTEIndex.LOWER_UNIVERSAL_TATTOO, AvatarTEIndex.SKIRT_TATTOO,
                             AvatarTEIndex.HAIR_TATTOO, AvatarTEIndex.EYES_TATTOO, AvatarTEIndex.LEFT_ARM_TATTOO,
                             AvatarTEIndex.LEFT_LEG_TATTOO, AvatarTEIndex.AUX1_TATTOO, AvatarTEIndex.AUX2_TATTOO,
                             AvatarTEIndex.AUX3_TATTOO),
    WearableType.PHYSICS: (),
}


@dataclasses.dataclass
class Wearable(SchemaBase):
    name: str
    wearable_type: WearableType
    permissions: InventoryPermissions
    sale_info: InventorySaleInfo
    # VisualParam ID -> val
    parameters: Dict[int, float]
    # TextureEntry ID -> texture UUID
    textures: Dict[int, UUID]

    @classmethod
    def _skip_to_next_populated_line(cls, reader: StringIO):
        old_pos = reader.tell()
        while peeked_data := reader.readline():
            # Read until we find a non-blank line
            if peeked_data.lstrip("\n"):
                break
            old_pos = reader.tell()
        # Reading an empty string means EOF
        if not peeked_data:
            raise SchemaParsingError("Premature EOF")
        reader.seek(old_pos)

    @classmethod
    def _read_and_parse_line(cls, reader: StringIO):
        cls._skip_to_next_populated_line(reader)
        return parse_schema_line(reader.readline())

    @classmethod
    def _read_expected_key(cls, reader: StringIO, expected_key: str) -> str:
        key, val = cls._read_and_parse_line(reader)
        if key != expected_key:
            raise ValueError(f"Expected {expected_key} not found, {(key, val)!r}")
        return val

    @classmethod
    def from_reader(cls, reader: StringIO) -> Wearable:
        cls._skip_to_next_populated_line(reader)
        version_str = reader.readline().rstrip()
        if version_str != WEARABLE_VERSION:
            raise ValueError(f"Bad wearable version {version_str!r}")
        cls._skip_to_next_populated_line(reader)
        name = reader.readline().rstrip()

        permissions = InventoryPermissions.from_reader(reader, read_header=True)
        sale_info = InventorySaleInfo.from_reader(reader, read_header=True)

        wearable_type = WearableType(int(cls._read_expected_key(reader, "type")))
        num_params = int(cls._read_expected_key(reader, "parameters"))
        params = {}
        for _ in range(num_params):
            param_id, param_val = cls._read_and_parse_line(reader)
            if param_val == ".":
                param_val = "0.0"
            params[int(param_id)] = float(param_val)

        num_textures = int(cls._read_expected_key(reader, "textures"))
        textures = {}
        for _ in range(num_textures):
            te_id, texture_id = cls._read_and_parse_line(reader)
            textures[int(te_id)] = UUID(texture_id)
        return Wearable(
            name=name,
            wearable_type=wearable_type,
            permissions=permissions,
            sale_info=sale_info,
            parameters=params,
            textures=textures
        )

    def to_writer(self, writer: StringIO):
        writer.write(f"{WEARABLE_VERSION}\n")
        writer.write(f"{self.name}\n\n")
        self.permissions.to_writer(writer)
        self.sale_info.to_writer(writer)
        writer.write(f"type {int(self.wearable_type)}\n")
        writer.write(f"parameters {len(self.parameters)}\n")
        for param_id, param_val in self.parameters.items():
            writer.write(f"{param_id} {param_val}\n")
        writer.write(f"textures {len(self.textures)}\n")
        for te_id, texture_id in self.textures.items():
            writer.write(f"{te_id} {texture_id}\n")

    @classmethod
    def make_default(cls, w_type: WearableType) -> Self:
        instance = cls(
            name="New " + w_type.name.replace("_", " ").title(),
            permissions=InventoryPermissions.make_default(),
            sale_info=InventorySaleInfo.make_default(),
            parameters={},
            textures={},
            wearable_type=w_type,
        )

        for te_idx in WEARABLE_TEXTURE_SLOTS[w_type]:
            instance.textures[te_idx] = DEFAULT_WEARABLE_TEX

        for param in VISUAL_PARAMS.by_wearable(w_type.name.lower()):
            instance.parameters[param.id] = param.value_default

        return instance



from __future__ import annotations

import asyncio
import sys
from pathlib import Path
from tempfile import TemporaryDirectory

from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.proxy import addon_ctx
from hippolyzer.lib.proxy.addon_utils import (
    BaseAddon,
    SessionProperty,
    send_chat,
    show_message,
)
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session
from hippolyzer.lib.proxy.test_utils import BaseProxyTest


class MockAddon(BaseAddon):
    bazquux: str = SessionProperty()
    another: str = SessionProperty("default")

    @handle_command(bar=str)
    async def foobar(self, _session: Session, _region: ProxiedRegion, bar: str):
        self.bazquux = bar
        self.another = bar
        send_chat(bar)
        show_message(bar)


PARENT_ADDON_SOURCE = """
from hippolyzer.lib.proxy.addon_utils import BaseAddon, GlobalProperty

class ParentAddon(BaseAddon):
    baz = None
    quux: int = GlobalProperty(0)

    @classmethod
    def foo(cls):
        cls.baz = 1

addons = [ParentAddon()]
"""

CHILD_ADDON_SOURCE = """
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.proxy.addons import AddonManager

import parent_addon

AddonManager.hot_reload(parent_addon)

class ChildAddon(BaseAddon):
    def handle_init(self, session_manager):
        parent_addon.ParentAddon.foo()

addons = [ChildAddon()]
"""


class AddonIntegrationTests(BaseProxyTest):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self.addon = MockAddon()
        AddonManager.init([], self.session_manager, [self.addon], swallow_addon_exceptions=False)
        self.temp_dir = TemporaryDirectory(prefix="addon_test_sources")
        self.child_path = Path(self.temp_dir.name) / "child_addon.py"
        self.parent_path = Path(self.temp_dir.name) / "parent_addon.py"

    def tearDown(self) -> None:
        AddonManager.shutdown()
        self.temp_dir.cleanup()

    def _fake_command(self, command: str) -> None:
        msg = Message(
            "ChatFromViewer",
            Block("AgentData", AgentID=self.session.agent_id, SessionID=self.session.id),
            Block("ChatData", Message=command, Channel=AddonManager.COMMAND_CHANNEL, fill_missing=True),
        )
        packet = self._msg_to_packet(msg, src=self.client_addr, dst=self.region_addr)
        self.protocol.handle_proxied_packet(packet)

    async def test_simple_command_setting_params(self):
        self._setup_default_circuit()
        self._fake_command("foobar baz")
        await self._wait_drained()
        self.assertEqual(self.session.addon_ctx["MockAddon"]["bazquux"], "baz")

        # In session context these should be equivalent
        with addon_ctx.push(new_session=self.session):
            self.assertEqual(self.session.addon_ctx["MockAddon"]["bazquux"], self.addon.bazquux)
            self.assertEqual(self.session.addon_ctx["MockAddon"]["another"], "baz")

        # Outside session context it should raise
        with self.assertRaises(AttributeError):
            self.assertEqual(self.addon.bazquux, "baz")
        # Even if there's a default value
        with self.assertRaises(AttributeError):
            self.assertEqual(self.addon.another, "baz")

        self.session.addon_ctx.clear()
        with addon_ctx.push(new_session=self.session):
            # This has no default so it should fail
            with self.assertRaises(AttributeError):
                _something = self.addon.bazquux
            # This has a default
            self.assertEqual(self.addon.another, "default")
        # Should have sent out the two injected packets for inbound and outbound chat
        # But not the original chatfromviewer from our command.
        self.assertEqual(len(self.transport.packets), 2)

    async def test_loading_addons(self):
        with open(self.parent_path, "w") as f:
            f.write(PARENT_ADDON_SOURCE)
        with open(self.child_path, "w") as f:
            f.write(CHILD_ADDON_SOURCE)
        AddonManager.load_addon_from_path(str(self.parent_path), reload=True)
        AddonManager.load_addon_from_path(str(self.child_path), reload=True)
        # Wait for the init hooks to run
        await asyncio.sleep(0.001)
        # Should be able to import this by name now
        import parent_addon  # noqa
        # ChildAddon calls a classmethod that mutates this
        self.assertEqual(1, parent_addon.ParentAddon.baz)

    async def test_unloading_addons(self):
        with open(self.parent_path, "w") as f:
            f.write(PARENT_ADDON_SOURCE)
        AddonManager.load_addon_from_path(str(self.parent_path), reload=True)
        # Wait for the init hooks to run
        await asyncio.sleep(0.001)
        # Should be able to import this by name now
        AddonManager.unload_addon_from_path(str(self.parent_path), reload=True)
        await asyncio.sleep(0.001)
        self.assertNotIn('hippolyzer.user_addon_parent_addon', sys.modules)

    async def test_global_property_access_and_set(self):
        with open(self.parent_path, "w") as f:
            f.write(PARENT_ADDON_SOURCE)
        AddonManager.load_addon_from_path(str(self.parent_path), reload=True)
        # Wait for the init hooks to run
        await asyncio.sleep(0.001)
        self.assertFalse("quux" in self.session_manager.addon_ctx["ParentAddon"])
        parent_addon_mod = AddonManager.FRESH_ADDON_MODULES['hippolyzer.user_addon_parent_addon']
        self.assertEqual(0, parent_addon_mod.ParentAddon.quux)
        self.assertEqual(0, self.session_manager.addon_ctx["ParentAddon"]["quux"])
        parent_addon_mod.ParentAddon.quux = 1
        self.assertEqual(1, self.session_manager.addon_ctx["ParentAddon"]["quux"])



import asyncio
from typing import Any, Optional, List, Tuple

from hippolyzer.lib.base.message.circuit import Circuit, ConnectionHolder
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.base.network.transport import AbstractUDPTransport, ADDR_TUPLE, UDPPacket


class MockTransport(AbstractUDPTransport):
    def sendto(self, data: Any, addr: Optional[ADDR_TUPLE] = ...) -> None:
        pass

    def abort(self) -> None:
        pass

    def close(self) -> None:
        pass

    def __init__(self):
        super().__init__()
        self.packets: List[Tuple[bytes, Tuple[str, int]]] = []

    def send_packet(self, packet: UDPPacket) -> None:
        self.packets.append((packet.data, packet.dst_addr))


class MockHandlingCircuit(Circuit):
    def __init__(self, handler: MessageHandler[Message, str]):
        super().__init__(("127.0.0.1", 1), ("127.0.0.1", 2), None)
        self.handler = handler

    def _send_prepared_message(self, message: Message, transport=None):
        loop = asyncio.get_event_loop_policy().get_event_loop()
        loop.call_soon(self.handler.handle, message)


class MockConnectionHolder(ConnectionHolder):
    def __init__(self, circuit, message_handler):
        self.circuit = circuit
        self.message_handler = message_handler


async def soon(awaitable) -> Message:
    return await asyncio.wait_for(awaitable, timeout=1.0)



"""
Example of how to make simple Caps requests
"""
import aiohttp

from hippolyzer.lib.proxy.addon_utils import BaseAddon, show_message
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class CapsExampleAddon(BaseAddon):
    @handle_command()
    async def test_caps(self, _session: Session, region: ProxiedRegion):
        caps_client = region.caps_client
        # We can pass in a ClientSession if we want to do keep-alive across requests
        async with aiohttp.ClientSession() as aio_sess:
            async with caps_client.get("SimulatorFeatures", session=aio_sess) as resp:
                await resp.read_llsd()
        # Or we can have one created for us just for this request
        async with caps_client.get("SimulatorFeatures") as resp:
            show_message(await resp.read_llsd())

        # POSTing LLSD works
        req = caps_client.post("AgentPreferences", llsd={
            "hover_height": 0.5,
        })
        # Request object can be built, then awaited
        async with req as resp:
            show_message(await resp.read_llsd())


addons = [CapsExampleAddon()]



"""
Example of proxy-provided caps

Useful for mocking out a cap that isn't actually implemented by the server
while developing the viewer-side pieces of it.

Implements a cap that accepts an `obj_id` UUID query parameter and returns
the name of the object.
"""
import asyncio
import asgiref.wsgi

from flask import Flask, Response, request

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.proxy import addon_ctx
from hippolyzer.lib.proxy.webapp_cap_addon import WebAppCapAddon

app = Flask("GetObjectNameCapApp")


@app.route('/')
async def get_object_name():
    # Should always have the current region, the cap handler is bound to one.
    # Just need to pull it from the `addon_ctx` module's global.
    obj_mgr = addon_ctx.region.get().objects
    obj_id = UUID(request.args['obj_id'])
    obj = obj_mgr.lookup_fullid(obj_id)
    if not obj:
        return Response(f"Couldn't find {obj_id!r}", status=404, mimetype="text/plain")

    try:
        await asyncio.wait_for(obj_mgr.request_object_properties(obj)[0], 1.0)
    except asyncio.TimeoutError:
        return Response(f"Timed out requesting {obj_id!r}'s properties", status=500, mimetype="text/plain")

    return Response(obj.Name, mimetype="text/plain")


class MockProxyCapExampleAddon(WebAppCapAddon):
    # A cap URL with this name will be tied to each region when
    # the sim is first connected to. The URL will be returned to the
    # viewer in the Seed if the viewer requests it by name.
    CAP_NAME = "GetObjectNameExample"
    # Any asgi app should be fine.
    APP = asgiref.wsgi.WsgiToAsgi(app)


addons = [MockProxyCapExampleAddon()]



from __future__ import annotations

import asyncio
import math
import multiprocessing
from urllib.parse import urlparse

import aioresponses
from mitmproxy.test import tflow, tutils
from mitmproxy.http import HTTPFlow, Headers
from yarl import URL

from hippolyzer.apps.proxy import run_http_proxy_process
from hippolyzer.lib.base.datatypes import Vector3
from hippolyzer.lib.base.helpers import create_logged_task
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.http_event_manager import MITMProxyEventManager
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.proxy.caps import SerializedCapData
from hippolyzer.lib.proxy.sessions import SessionManager
from hippolyzer.lib.proxy.test_utils import BaseProxyTest


class MockAddon(BaseAddon):
    def handle_http_request(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        flow.metadata["touched_addon"] = True

    def handle_http_response(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        flow.metadata["touched_addon"] = True


class HTTPIntegrationTests(BaseProxyTest):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self.addon = MockAddon()
        AddonManager.init([], self.session_manager, [self.addon])
        self.flow_context = self.session_manager.flow_context
        self.http_event_manager = MITMProxyEventManager(self.session_manager, self.flow_context)
        self._setup_default_circuit()

    async def _pump_one_event(self):
        # If we don't yield then the new entry won't end up in the queue
        await asyncio.sleep(0.001)
        await self.http_event_manager.pump_proxy_event()
        await asyncio.sleep(0.001)

    async def test_http_flow_request(self):
        # mimic a request coming in from mitmproxy over the queue
        fake_flow = tflow.tflow(req=tutils.treq(host="example.com"))
        fake_flow.metadata["cap_data_ser"] = SerializedCapData()
        self.flow_context.from_proxy_queue.put(("request", fake_flow.get_state()), True)
        await self._pump_one_event()
        self.assertTrue(self.flow_context.from_proxy_queue.empty())
        self.assertFalse(self.flow_context.to_proxy_queue.empty())
        flow_state = self.flow_context.to_proxy_queue.get(True)[2]
        mitm_flow: HTTPFlow = HTTPFlow.from_state(flow_state)
        # The response sent back to mitmproxy should have been our modified version
        self.assertEqual(True, mitm_flow.metadata["touched_addon"])

    async def test_http_flow_response(self):
        # mimic a request coming in from mitmproxy over the queue
        fake_flow = tflow.tflow(req=tutils.treq(host="example.com"), resp=tutils.tresp())
        fake_flow.metadata["cap_data_ser"] = SerializedCapData()
        self.flow_context.from_proxy_queue.put(("response", fake_flow.get_state()), True)
        await self._pump_one_event()
        self.assertTrue(self.flow_context.from_proxy_queue.empty())
        self.assertFalse(self.flow_context.to_proxy_queue.empty())
        flow_state = self.flow_context.to_proxy_queue.get(True)[2]
        mitm_flow: HTTPFlow = HTTPFlow.from_state(flow_state)
        # The response sent back to mitmproxy should have been our modified version
        self.assertEqual(True, mitm_flow.metadata["touched_addon"])

    async def test_firestorm_bridge_avatar_z_pos(self):
        # Simulate an avatar with a non-finite Z pos in a coarselocation
        self.session.main_region.objects.state.coarse_locations.update({
            self.session.agent_id: Vector3(1, 2, math.inf),
        })
        self.session.objects._rebuild_avatar_objects()
        # GuessedZ should be picked up for the avatar based on the bridge request
        fake_flow = tflow.tflow(
            req=tutils.treq(host="example.com", content=b'<llsd><string>getZOffsets|'),
            resp=tutils.tresp(
                headers=Headers((
                    (b"X-SecondLife-Object-Name", b"#Firestorm LSL Bridge v99999"),
                    (b"X-SecondLife-Owner-Key", str(self.session.agent_id).encode("utf8")),
                )),
                content=f"<llsd><string>{self.session.agent_id}, 2000.0</string></llsd>".encode("utf8")
            ),
        )
        fake_flow.metadata["cap_data_ser"] = SerializedCapData("FirestormBridge")
        fake_flow.metadata["from_browser"] = False
        self.session_manager.flow_context.from_proxy_queue.put(("response", fake_flow.get_state()), True)
        await self._pump_one_event()
        av = tuple(self.session.objects.all_avatars)[0]
        self.assertEqual(Vector3(1, 2, 2000.0), av.RegionPosition)

    async def test_asset_server_proxy_wrapper_caps(self):
        # We support "wrapper caps" that disambiguate otherwise ambiguous caps.
        # The URL provided by the sim may not be unique across regions or sessions,
        # in the case of ViewerAsset on agni, so we generate a random hostname
        # as an alias and send that to the viewer instead.
        region = self.session.main_region
        region.update_caps({
            "ViewerAsset": "http://assets.local/foo",
        })
        wrapper_url = region.register_wrapper_cap("ViewerAsset")
        parsed = urlparse(wrapper_url)
        fake_flow = tflow.tflow(req=tutils.treq(
            host=parsed.hostname,
            path="/foo/baz?asset_id=bar",
            port=80,
        ))
        fake_flow.metadata["cap_data_ser"] = SerializedCapData()
        self.flow_context.from_proxy_queue.put(("request", fake_flow.get_state()), True)
        await self._pump_one_event()
        flow_state = self.flow_context.to_proxy_queue.get(True)[2]
        mitm_flow: HTTPFlow = HTTPFlow.from_state(flow_state)
        self.assertIsNotNone(mitm_flow.response)
        self.assertEqual(307, mitm_flow.response.status_code)
        self.assertEqual(
            "http://assets.local/foo/baz?asset_id=bar",
            mitm_flow.response.headers["Location"],
        )


class TestCapsClient(BaseProxyTest):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self._setup_default_circuit()
        self.caps_client = self.session.main_region.caps_client

    async def test_requests_proxied_by_default(self):
        with aioresponses.aioresponses() as m:
            m.get("http://example.com/", body=b"foo")
            async with self.caps_client.get("http://example.com/") as resp:
                self.assertEqual(await resp.read(), b"foo")
            kwargs = m.requests[("GET", URL("http://example.com/"))][0].kwargs
        # Request should have been proxied, with a header marking it
        self.assertEqual(kwargs['headers']["X-Hippo-Injected"], "1")
        self.assertEqual(kwargs['proxy'], "http://127.0.0.1:9062")


class TestMITMProxy(BaseProxyTest):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self._setup_default_circuit()
        self.caps_client = self.session.main_region.caps_client
        proxy_port = 9905
        self.session_manager.settings.HTTP_PROXY_PORT = proxy_port

        self.http_proc = multiprocessing.Process(
            target=run_http_proxy_process,
            args=("127.0.0.1", proxy_port, self.session_manager.flow_context),
            daemon=True,
        )
        self.http_proc.start()
        self.session_manager.flow_context.mitmproxy_ready.wait(1.0)

        self.http_event_manager = MITMProxyEventManager(
            self.session_manager,
            self.session_manager.flow_context
        )

    def test_mitmproxy_works(self):
        async def _request_example_com():
            # Pump callbacks from mitmproxy
            _ = create_logged_task(self.http_event_manager.run())
            try:
                async with self.caps_client.get("http://example.com/", timeout=0.5) as resp:
                    self.assertIn(b"Example Domain", await resp.read())
                async with self.caps_client.get("https://example.com/", timeout=0.5) as resp:
                    self.assertIn(b"Example Domain", await resp.read())
            finally:
                # Tell the event pump and mitmproxy they need to shut down
                self.session_manager.flow_context.shutdown_signal.set()
        asyncio.run(_request_example_com())
        self.http_proc.join()



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import logging
import unittest
import binascii

from hippolyzer.lib.base.message.udpserializer import UDPMessageSerializer
from hippolyzer.lib.base.settings import Settings
from hippolyzer.lib.base.message.data import msg_tmpl
from hippolyzer.lib.base.message.udpdeserializer import UDPMessageDeserializer
from hippolyzer.lib.base.message.template_parser import MessageTemplateParser
from hippolyzer.lib.base.message.template_dict import TemplateDictionary

AGENT_DATA_UPDATE = "C0 00 00 00 02 00 FF FF 01 83 1C 8A 77 67 E3 7B 42 2E AF B3 85 09 31 97 CA D1 03 4A 42 00 01 06 " \
                    "4B 72 61 66 74 00 01 01 00 1A "
AGENT_DATA_UPDATE = binascii.unhexlify(''.join(AGENT_DATA_UPDATE.split()))

AGENT_ANIMATION = "40 00 00 00 0A 00 05 1C 8A 77 67 E3 7B 42 2E AF B3 85 09 31 97 CA D1 4B 6F FF 1F B5 67 41 FD 85 EF" \
                  " A1 98 3B F2 B5 77 01 EF CF 67 0C 2D 18 81 28 97 3A 03 4E BC 80 6B 67 00 01 00 "
AGENT_ANIMATION = binascii.unhexlify(''.join(AGENT_ANIMATION.split()))

OBJECT_UPDATE = "C0 00 00 00 51 00 0C 00 01 EA 03 00 02 E6 03 00 01 BE FF 01 06 BC 8E 0B 00 01 69 94 8C 6A 4D 22 1B " \
                "66 EC E4 AC 31 63 93 CB 4B 57 89 98 01 09 03 00 01 51 40 88 3E 51 40 88 3E 51 40 88 3E 3C A2 44 B3 " \
                "42 9A 68 2B 42 C8 5B D1 41 00 18 4B 8C FF 3E BD 76 FF BE C5 44 00 01 BF 00 10 50 04 00 01 10 20 05 " \
                "00 04 64 64 00 0F 2E 00 01 A1 33 DC 77 0A 1A BB 27 A7 2E 78 64 63 AB 94 AB 00 08 80 3F 00 03 80 3F " \
                "00 0F 10 13 FF 00 08 80 3F 8F C2 F5 3D 00 0A 56 6F 43 CC 00 01 02 00 03 04 00 02 04 00 02 64 26 00 " \
                "03 0E 00 01 0E 00 01 19 00 01 80 00 01 80 00 01 80 00 01 80 00 01 80 00 01 80 91 11 D2 5E 2F 12 8F " \
                "81 55 A7 40 3A 78 B3 0E 2D 00 10 03 01 00 03 1E 25 6E A2 FF C5 E0 83 00 01 06 00 01 0D 0D 01 00 11 " \
                "0E DC 9B 83 98 9A 4A 76 AC C3 DB BF 37 54 61 88 00 22 "
OBJECT_UPDATE = binascii.unhexlify(''.join(OBJECT_UPDATE.split()))

COARSE_LOCATION_UPDATE = b'\x00\x00\x00\x00E\x00\xff\x06\x00\xff\xff\xff\xff\x00'

UNKNOWN_PACKET = b'\x00\x00\x00\x00E\x00\xff\xf0\x00\xff\xff\xff\xff\x00'


class TestPacketDecode(unittest.TestCase):

    def tearDown(self):
        pass

    def setUp(self):
        self.maxDiff = None
        self.template_file = msg_tmpl
        parser = MessageTemplateParser(self.template_file)
        self.template_list = parser.message_templates
        self.template_dict = TemplateDictionary(self.template_list)

        self.settings = Settings()
        self.settings.ENABLE_DEFERRED_PACKET_PARSING = False

    def test_agent_data_update(self):
        """test if the agent data update packet can be decoded"""
        message = AGENT_DATA_UPDATE
        deserializer = UDPMessageDeserializer(settings=self.settings)
        packet = deserializer.deserialize(message)
        assert packet is not None, "Wrong data"
        assert packet.name == 'AgentDataUpdate', "Wrong packet"
        assert packet.blocks['AgentData'][0].vars['LastName'] == 'Kraft', \
            "Wrong last name for agent update"
        assert packet.blocks['AgentData'][0].vars['FirstName'] == 'JB', \
            "Wrong first name for agent update"

    def test_agent_animation(self):
        """test if the agent data update packet can be decoded"""
        message = AGENT_ANIMATION
        deserializer = UDPMessageDeserializer(settings=self.settings)
        packet = deserializer.deserialize(message)
        assert packet is not None, "Wrong data 2"

    def test_object_update(self):
        """test if the object update packet can be decoded"""

        message = OBJECT_UPDATE

        deserializer = UDPMessageDeserializer(settings=self.settings)
        packet = deserializer.deserialize(message)
        assert packet is not None, "Wrong data"
        assert packet.name == 'ObjectUpdate', "Wrong packet"

        blocks = {}

        for block in packet.blocks:
            blocks[block] = 1

        self.assertTrue("RegionData" in blocks)
        self.assertTrue("ObjectData" in blocks)

    def test_course_location_roundtrips(self):
        message = COARSE_LOCATION_UPDATE
        deserializer = UDPMessageDeserializer(settings=self.settings)
        serializer = UDPMessageSerializer()
        parsed = deserializer.deserialize(message)
        logging.debug("Parsed blocks: %r " % (list(parsed.blocks.keys()),))
        self.assertEqual(message, serializer.serialize(parsed))

    def test_unknown_packet_roundtrips(self):
        message = UNKNOWN_PACKET
        deserializer = UDPMessageDeserializer(settings=self.settings)
        serializer = UDPMessageSerializer()
        parsed = deserializer.deserialize(message)
        logging.debug("Parsed blocks: %r " % (list(parsed.blocks.keys()),))
        self.assertEqual("UnknownMessage:240", parsed.name)
        self.assertEqual(message, serializer.serialize(parsed))
